from __future__ import annotations

import sys
from pathlib import Path

src_path = Path(__file__).parent.parent / "src"
if str(src_path) not in sys.path:
    sys.path.insert(0, str(src_path))


def sphere_objective(x):
    import numpy as np

    return -np.sum(x**2, axis=1)


def make_from_unit_fn(bounds):
    from enn.turbo_utils import from_unit

    def from_unit_fn(x):
        return from_unit(x, bounds)

    return from_unit_fn


def make_fallback_fn(bounds, rng):
    from enn.turbo_utils import from_unit

    def fallback_fn(x, n):
        idx = rng.choice(x.shape[0], size=n, replace=False)
        return from_unit(x[idx], bounds)

    return fallback_fn


def make_select_sobol_fn(bounds, rng):
    from enn.turbo_utils import from_unit

    def select_sobol_fn(x, n):
        idx = rng.choice(x.shape[0], size=n, replace=False)
        return from_unit(x[idx], bounds)

    return select_sobol_fn



from .core import EpistemicNearestNeighbors
from .fit import enn_fit
from .turbo import Turbo, TurboMode

__all__: list[str] = [
    "EpistemicNearestNeighbors",
    "TurboMode",
    "Turbo",
    "enn_fit",
]



from __future__ import annotations

import pytest


def test_ennnormal_sample_shape_and_clip():
    import numpy as np

    from enn.enn_normal import ENNNormal

    rng = np.random.default_rng(0)
    mu = np.array([[0.0, 1.0]], dtype=float)
    se = np.array([[1.0, 2.0]], dtype=float)
    normal = ENNNormal(mu=mu, se=se)
    samples = normal.sample(5, clip=1.0, rng=rng)
    assert samples.shape == (1, 2, 5)
    assert np.all(samples >= mu.min() - 2.0)
    assert np.all(samples <= mu.max() + 2.0)


def test_epistemic_nearest_neighbors_posterior_and_var_scale_and_hnsw_threshold():
    import numpy as np

    from enn.core import EpistemicNearestNeighbors
    from enn.enn_params import ENNParams

    rng = np.random.default_rng(0)
    n = 20
    d = 3
    x = rng.standard_normal((n, d))
    y = (x.sum(axis=1, keepdims=True)).astype(float)
    yvar = 0.1 * np.ones_like(y)
    model_flat = EpistemicNearestNeighbors(x, y, yvar, hnsw_threshold=None)
    model_hnsw = EpistemicNearestNeighbors(x, y, yvar, hnsw_threshold=5)
    x_test = rng.standard_normal((4, d))
    params = ENNParams(k=3, var_scale=1.0)
    post_flat = model_flat.posterior(x_test, params=params, exclude_nearest=False)
    post_hnsw = model_hnsw.posterior(x_test, params=params, exclude_nearest=True)
    assert post_flat.mu.shape == (4, 1)
    assert post_flat.se.shape == (4, 1)
    assert post_hnsw.mu.shape == (4, 1)
    assert post_hnsw.se.shape == (4, 1)
    post_changed = model_flat.posterior(
        x_test, params=ENNParams(k=5, var_scale=0.5), exclude_nearest=True
    )
    assert post_changed.mu.shape == (4, 1)
    assert post_changed.se.shape == (4, 1)


def test_epistemic_nearest_neighbors_with_no_observations_returns_prior_like_posterior():
    import numpy as np

    from enn.core import EpistemicNearestNeighbors
    from enn.enn_params import ENNParams

    rng = np.random.default_rng(0)
    d = 3
    x = np.zeros((0, d), dtype=float)
    y = np.zeros((0, 1), dtype=float)
    yvar = np.ones_like(y, dtype=float)
    model = EpistemicNearestNeighbors(x, y, yvar, hnsw_threshold=None)
    x_test = rng.standard_normal((5, d))
    post = model.posterior(
        x_test, params=ENNParams(k=3, var_scale=1.0), exclude_nearest=False
    )
    assert post.mu.shape == (5, 1)
    assert post.se.shape == (5, 1)
    assert np.allclose(post.mu, 0.0)
    assert np.allclose(post.se, 1.0)


@pytest.mark.parametrize("num_obs", [1, 2, 3])
def test_epistemic_nearest_neighbors_with_few_observations_has_valid_posterior(
    num_obs: int,
):
    import numpy as np

    from enn.core import EpistemicNearestNeighbors
    from enn.enn_params import ENNParams

    rng = np.random.default_rng(0)
    d = 3
    x = rng.standard_normal((num_obs, d))
    y = (x.sum(axis=1, keepdims=True)).astype(float)
    yvar = 0.1 * np.ones_like(y)
    model = EpistemicNearestNeighbors(x, y, yvar, hnsw_threshold=None)
    x_test = rng.standard_normal((5, d))
    post = model.posterior(
        x_test, params=ENNParams(k=3, var_scale=1.0), exclude_nearest=False
    )
    assert post.mu.shape == (5, 1)
    assert post.se.shape == (5, 1)
    assert np.all(np.isfinite(post.mu))
    assert np.all(np.isfinite(post.se))


def test_batch_posterior_matches_individual_posterior_calls():
    import numpy as np

    from enn.core import EpistemicNearestNeighbors
    from enn.enn_params import ENNParams

    rng = np.random.default_rng(0)
    n = 20
    d = 3
    x = rng.standard_normal((n, d))
    y = (x.sum(axis=1, keepdims=True)).astype(float)
    yvar = 0.1 * np.ones_like(y)
    model = EpistemicNearestNeighbors(x, y, yvar, hnsw_threshold=None)
    x_test = rng.standard_normal((4, d))
    paramss = [
        ENNParams(k=3, var_scale=1.0),
        ENNParams(k=5, var_scale=0.5),
        ENNParams(k=7, var_scale=2.0),
    ]
    post_batch = model.batch_posterior(x_test, paramss, exclude_nearest=False)
    assert post_batch.mu.shape == (len(paramss), x_test.shape[0], model.num_outputs)
    assert post_batch.se.shape == (len(paramss), x_test.shape[0], model.num_outputs)
    for i, params in enumerate(paramss):
        post = model.posterior(x_test, params=params, exclude_nearest=False)
        assert np.allclose(post_batch.mu[i], post.mu)
        assert np.allclose(post_batch.se[i], post.se)


def test_batch_posterior_matches_individual_posterior_calls_with_exclude_nearest():
    import numpy as np

    from enn.core import EpistemicNearestNeighbors
    from enn.enn_params import ENNParams

    rng = np.random.default_rng(0)
    n = 20
    d = 3
    x = rng.standard_normal((n, d))
    y = (x.sum(axis=1, keepdims=True)).astype(float)
    yvar = 0.1 * np.ones_like(y)
    model = EpistemicNearestNeighbors(x, y, yvar, hnsw_threshold=None)
    x_test = rng.standard_normal((4, d))
    paramss = [
        ENNParams(k=3, var_scale=1.0),
        ENNParams(k=5, var_scale=0.5),
    ]
    post_batch = model.batch_posterior(x_test, paramss, exclude_nearest=True)
    assert post_batch.mu.shape == (len(paramss), x_test.shape[0], model.num_outputs)
    assert post_batch.se.shape == (len(paramss), x_test.shape[0], model.num_outputs)
    for i, params in enumerate(paramss):
        post = model.posterior(x_test, params=params, exclude_nearest=True)
        assert np.allclose(post_batch.mu[i], post.mu)
        assert np.allclose(post_batch.se[i], post.se)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import numpy as np
    from gpytorch.likelihoods import GaussianLikelihood

    from .turbo_gp import TurboGP


def standardize_y(y) -> tuple[float, float]:
    import numpy as np

    y_array = np.asarray(y, dtype=float)
    center = float(np.median(y_array))
    scale = float(np.std(y_array))
    if not np.isfinite(scale) or scale <= 0.0:
        scale = 1.0
    return center, scale


def fit_gp(
    x_obs_list: list,
    y_obs_list: list,
    num_dim: int,
    *,
    num_steps: int = 50,
) -> tuple[
    "TurboGP | None",
    "GaussianLikelihood | None",
    float,
    float,
]:
    import numpy as np
    import torch
    from gpytorch.constraints import Interval
    from gpytorch.likelihoods import GaussianLikelihood
    from gpytorch.mlls import ExactMarginalLogLikelihood

    from .turbo_gp import TurboGP

    x = np.asarray(x_obs_list, dtype=float)
    y = np.asarray(y_obs_list, dtype=float)
    n = x.shape[0]
    if n == 0:
        return None, None, 0.0, 1.0
    if n == 1:
        gp_y_mean = float(y[0])
        gp_y_std = 1.0
        return None, None, gp_y_mean, gp_y_std
    gp_y_mean, gp_y_std = standardize_y(y)
    y_centered = y - gp_y_mean
    z = y_centered / gp_y_std
    train_x = torch.as_tensor(x, dtype=torch.float64)
    train_y = torch.as_tensor(z, dtype=torch.float64)
    noise_constraint = Interval(5e-4, 0.2)
    lengthscale_constraint = Interval(0.005, 2.0)
    outputscale_constraint = Interval(0.05, 20.0)
    likelihood = GaussianLikelihood(noise_constraint=noise_constraint).to(
        dtype=train_y.dtype
    )
    model = TurboGP(
        train_x=train_x,
        train_y=train_y,
        likelihood=likelihood,
        lengthscale_constraint=lengthscale_constraint,
        outputscale_constraint=outputscale_constraint,
        ard_dims=num_dim,
    ).to(dtype=train_x.dtype)
    model.covar_module.outputscale = torch.tensor(1.0, dtype=train_x.dtype)
    model.covar_module.base_kernel.lengthscale = torch.full(
        (num_dim,), 0.5, dtype=train_x.dtype
    )
    likelihood.noise = torch.tensor(0.005, dtype=train_y.dtype)
    model.train()
    likelihood.train()
    mll = ExactMarginalLogLikelihood(likelihood, model)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.1)
    for _ in range(num_steps):
        optimizer.zero_grad()
        output = model(train_x)
        loss = -mll(output, train_y)
        loss.backward()
        optimizer.step()
    model.eval()
    likelihood.eval()
    return model, likelihood, gp_y_mean, gp_y_std


def latin_hypercube(num_points: int, num_dim: int, *, rng) -> np.ndarray:
    import numpy as np

    x = np.zeros((num_points, num_dim))
    centers = (1.0 + 2.0 * np.arange(0.0, num_points)) / float(2 * num_points)
    for j in range(num_dim):
        x[:, j] = centers[rng.permutation(num_points)]
    pert = rng.uniform(-1.0, 1.0, size=(num_points, num_dim)) / float(2 * num_points)
    x += pert
    return x


def argmax_random_tie(values, *, rng) -> int:
    import numpy as np

    if values.ndim != 1:
        raise ValueError(values.shape)
    max_val = float(np.max(values))
    idx = np.nonzero(values >= max_val)[0]
    if idx.size == 0:
        return int(rng.integers(values.size))
    if idx.size == 1:
        return int(idx[0])
    j = int(rng.integers(idx.size))
    return int(idx[j])


def pareto_front(mu, se) -> np.ndarray:
    import numpy as np

    if mu.shape != se.shape or mu.ndim != 1:
        raise ValueError((mu.shape, se.shape))
    n = mu.size
    if n == 0:
        return np.zeros((0,), dtype=bool)
    order = np.argsort(-mu)
    mu_sorted = mu[order]
    se_sorted = se[order]
    is_pareto_sorted = np.zeros_like(mu_sorted, dtype=bool)
    best_se = float("inf")
    for i in range(n):
        if se_sorted[i] < best_se:
            best_se = float(se_sorted[i])
            is_pareto_sorted[i] = True
    is_pareto = np.zeros_like(is_pareto_sorted, dtype=bool)
    is_pareto[order] = is_pareto_sorted
    return is_pareto


def arms_from_pareto_fronts(x_cand, mu, se, num_arms, rng) -> np.ndarray:
    import numpy as np

    if x_cand.ndim != 2:
        raise ValueError(x_cand.shape)
    if mu.shape != se.shape or mu.ndim != 1:
        raise ValueError((mu.shape, se.shape))
    if mu.size != x_cand.shape[0]:
        raise ValueError((mu.size, x_cand.shape[0]))

    mu_neg = -mu

    i = np.argsort(mu_neg)
    x_cand_sorted = x_cand[i]
    se_sorted = se[i]

    i_all = list(range(len(se_sorted)))
    i_keep = []

    while len(i_keep) < num_arms:
        se_max = float("-inf")
        i_front = []
        for idx in i_all:
            if se_sorted[idx] >= se_max:
                i_front.append(idx)
                se_max = se_sorted[idx]
        if len(i_keep) + len(i_front) <= num_arms:
            i_keep.extend(i_front)
        else:
            i_keep.extend(
                rng.choice(i_front, size=num_arms - len(i_keep), replace=False)
            )
        i_all = sorted(set(i_all) - set(i_front))

    i_keep = np.array(i_keep)
    x_arms = x_cand_sorted[i_keep]

    return x_arms


def sobol_perturb_np(
    x_center, lb, ub, num_candidates, mask, *, sobol_engine
) -> np.ndarray:
    import numpy as np

    sobol_samples = sobol_engine.random(num_candidates)
    lb_array = np.asarray(lb)
    ub_array = np.asarray(ub)
    pert = lb_array + (ub_array - lb_array) * sobol_samples
    candidates = np.tile(x_center, (num_candidates, 1))
    if np.any(mask):
        candidates[mask] = pert[mask]
    return candidates


def raasp(
    x_center, lb, ub, num_candidates, *, num_pert: int = 20, rng, sobol_engine
) -> np.ndarray:
    import numpy as np

    num_dim = x_center.shape[-1]
    prob_perturb = min(num_pert / num_dim, 1.0)
    mask = rng.random((num_candidates, num_dim)) <= prob_perturb
    ind = np.where(np.sum(mask, axis=1) == 0)[0]
    if len(ind) > 0:
        mask[ind, rng.integers(0, num_dim, size=len(ind))] = True
    return sobol_perturb_np(
        x_center, lb, ub, num_candidates, mask, sobol_engine=sobol_engine
    )


def to_unit(x, bounds) -> np.ndarray:
    import numpy as np

    lb = bounds[:, 0]
    ub = bounds[:, 1]
    if np.any(ub <= lb):
        raise ValueError(bounds)
    return (x - lb) / (ub - lb)


def from_unit(x_unit, bounds) -> np.ndarray:
    lb = bounds[:, 0]
    ub = bounds[:, 1]
    return lb + x_unit * (ub - lb)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gpytorch.distributions import MultivariateNormal


def _get_exact_gp_base():
    from gpytorch.models import ExactGP

    return ExactGP


class TurboGP(_get_exact_gp_base()):
    def __init__(
        self,
        train_x,
        train_y,
        likelihood,
        lengthscale_constraint,
        outputscale_constraint,
        ard_dims: int,
    ) -> None:
        from gpytorch.kernels import MaternKernel, ScaleKernel
        from gpytorch.means import ConstantMean

        super().__init__(train_x, train_y, likelihood)
        self.mean_module = ConstantMean()
        base_kernel = MaternKernel(
            nu=2.5,
            ard_num_dims=ard_dims,
            lengthscale_constraint=lengthscale_constraint,
        )
        self.covar_module = ScaleKernel(
            base_kernel,
            outputscale_constraint=outputscale_constraint,
        )

    def forward(self, x) -> MultivariateNormal:
        from gpytorch.distributions import MultivariateNormal

        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)

    def posterior(self, x) -> MultivariateNormal:
        return self(x)



def test_subsample_loglik_and_enn_fit_improve_hyperparameters():
    import numpy as np

    from enn.core import EpistemicNearestNeighbors
    from enn.enn_params import ENNParams
    from enn.fit import enn_fit, subsample_loglik

    rng = np.random.default_rng(0)
    n = 40
    d = 2
    x = rng.standard_normal((n, d))
    true_w = np.array([1.5, -0.5])
    y_mean = x @ true_w
    noise_std = 0.1
    noise = noise_std * rng.standard_normal(n)
    y = (y_mean + noise).reshape(-1, 1)
    yvar = (noise_std**2) * np.ones_like(y)
    model = EpistemicNearestNeighbors(x, y, yvar, hnsw_threshold=None)
    rng_fit = np.random.default_rng(1)
    result = enn_fit(
        model,
        num_fit_candidates=30,
        num_fit_samples=20,
        rng=rng_fit,
    )
    assert isinstance(result, ENNParams)
    assert result.k >= 3, "k must be >= 3 per style guide"
    assert result.var_scale > 0.0
    rng_eval = np.random.default_rng(2)
    tuned_lls = subsample_loglik(
        model,
        x,
        y[:, 0],
        paramss=[result],
        P=20,
        rng=rng_eval,
    )
    tuned_ll = tuned_lls[0]
    assert np.isfinite(tuned_ll), "tuned log-likelihood must be finite"



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Optional

if TYPE_CHECKING:
    import numpy as np

    from .turbo_gp import TurboGP

from .turbo_utils import standardize_y


def select_enn_pareto(
    x_cand: np.ndarray,
    num_arms: int,
    x_obs_list: list,
    y_obs_list: list,
    k: Optional[int],
    var_scale: float,
    hnsw_threshold: Optional[int],
    rng,
    fallback_fn: Callable[[np.ndarray, int], np.ndarray],
    from_unit_fn: Callable[[np.ndarray], np.ndarray],
) -> np.ndarray:
    import numpy as np

    from .core import EpistemicNearestNeighbors
    from .enn_params import ENNParams
    from .turbo_utils import arms_from_pareto_fronts

    if len(x_obs_list) == 0:
        return fallback_fn(x_cand, num_arms)
    y_obs_array = np.asarray(y_obs_list, dtype=float)
    if y_obs_array.size == 0:
        return fallback_fn(x_cand, num_arms)

    mu_y, sigma_y = standardize_y(y_obs_array)
    y_standardized = (y_obs_array - mu_y) / sigma_y

    y = y_standardized.reshape(-1, 1)
    yvar = np.zeros_like(y, dtype=float)
    x_obs_array = np.asarray(x_obs_list, dtype=float)
    enn_model = EpistemicNearestNeighbors(
        x_obs_array,
        y,
        yvar,
        hnsw_threshold=hnsw_threshold,
    )
    if len(enn_model) == 0:
        return fallback_fn(x_cand, num_arms)

    if k is None:
        k = 10

    params = ENNParams(k=k, var_scale=var_scale)
    posterior = enn_model.posterior(x_cand, params=params)
    mu = posterior.mu[:, 0]
    se = posterior.se[:, 0]

    x_arms = arms_from_pareto_fronts(x_cand, mu, se, num_arms, rng)
    return from_unit_fn(x_arms)


def select_uniform(
    x_cand: np.ndarray,
    num_arms: int,
    num_dim: int,
    rng,
    from_unit_fn: Callable[[np.ndarray], np.ndarray],
) -> np.ndarray:
    if x_cand.ndim != 2 or x_cand.shape[1] != num_dim:
        raise ValueError(x_cand.shape)
    if x_cand.shape[0] < num_arms:
        raise ValueError((x_cand.shape[0], num_arms))
    idx = rng.choice(x_cand.shape[0], size=num_arms, replace=False)
    return from_unit_fn(x_cand[idx])


def select_gp_thompson(
    x_cand: np.ndarray,
    num_arms: int,
    x_obs_list: list,
    y_obs_list: list,
    num_dim: int,
    gp_num_steps: int,
    rng,
    gp_y_mean: float,
    gp_y_std: float,
    select_sobol_fn: Callable[[np.ndarray, int], np.ndarray],
    from_unit_fn: Callable[[np.ndarray], np.ndarray],
    *,
    model: Optional["TurboGP"] = None,
    new_gp_y_mean: Optional[float] = None,
    new_gp_y_std: Optional[float] = None,
) -> tuple[np.ndarray, float, float, TurboGP | None]:
    import contextlib

    import gpytorch
    import numpy as np
    import torch

    from .turbo_utils import fit_gp

    @contextlib.contextmanager
    def _torch_rng_context(generator):
        old_state = torch.get_rng_state()
        try:
            torch.set_rng_state(generator.get_state())
            yield
        finally:
            torch.set_rng_state(old_state)

    if len(x_obs_list) == 0:
        return select_sobol_fn(x_cand, num_arms), gp_y_mean, gp_y_std, None
    if model is None:
        model, _likelihood, new_gp_y_mean, new_gp_y_std = fit_gp(
            x_obs_list,
            y_obs_list,
            num_dim,
            num_steps=gp_num_steps,
        )
    if model is None:
        return select_sobol_fn(x_cand, num_arms), gp_y_mean, gp_y_std, None
    if new_gp_y_mean is None:
        new_gp_y_mean = gp_y_mean
    if new_gp_y_std is None:
        new_gp_y_std = gp_y_std
    x_torch = torch.as_tensor(x_cand, dtype=torch.float64)
    seed = int(rng.integers(2**31 - 1))
    gen = torch.Generator(device=x_torch.device)
    gen.manual_seed(seed)
    with (
        torch.no_grad(),
        gpytorch.settings.fast_pred_var(),
        _torch_rng_context(gen),
    ):
        posterior = model.posterior(x_torch)
        samples = posterior.sample(
            sample_shape=torch.Size([1]),
        )
    ts = samples[0].reshape(-1)
    scores = ts.detach().cpu().numpy().reshape(-1)
    scores = new_gp_y_mean + new_gp_y_std * scores
    if x_cand.shape[0] < num_arms:
        raise ValueError((x_cand.shape[0], num_arms))
    idx = np.argpartition(-scores, num_arms - 1)[:num_arms]
    return from_unit_fn(x_cand[idx]), new_gp_y_mean, new_gp_y_std, model



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class TrustRegionState:
    num_dim: int
    num_arms: int
    length: float = 0.8
    length_init: float = 0.8
    length_min: float = 0.5**7
    length_max: float = 1.6
    failure_counter: int = 0
    success_counter: int = 0
    best_value: float = -float("inf")
    prev_num_obs: int = 0

    def __post_init__(self) -> None:
        import numpy as np

        self.failure_tolerance = int(
            np.ceil(
                max(
                    4.0 / float(self.num_arms),
                    float(self.num_dim) / float(self.num_arms),
                )
            )
        )
        self.success_tolerance = 3

    def update(self, values) -> None:
        import numpy as np

        if values.ndim != 1:
            raise ValueError(values.shape)
        if values.size == 0:
            return
        new_values = values[self.prev_num_obs :]
        if new_values.size == 0:
            return
        if not np.isfinite(self.best_value):
            self.best_value = float(np.max(new_values))
            self.prev_num_obs = values.size
            return
        improved = np.max(new_values) > self.best_value + 1e-3 * np.abs(self.best_value)
        if improved:
            self.success_counter += 1
            self.failure_counter = 0
        else:
            self.success_counter = 0
            self.failure_counter += 1
        if self.success_counter >= self.success_tolerance:
            self.length = min(2.0 * self.length, self.length_max)
            self.success_counter = 0
        elif self.failure_counter >= self.failure_tolerance:
            self.length = 0.5 * self.length
            self.failure_counter = 0
        self.best_value = max(self.best_value, float(np.max(new_values)))
        self.prev_num_obs = values.size

    def needs_restart(self) -> bool:
        return self.length < self.length_min

    def restart(self) -> None:
        self.length = self.length_init
        self.failure_counter = 0
        self.success_counter = 0
        self.best_value = -float("inf")
        self.prev_num_obs = 0

    def _compute_bounds_1d(self, x_center, weights=None):
        import numpy as np

        if x_center.ndim != 1 or x_center.shape[0] != self.num_dim:
            raise ValueError(x_center.shape)
        if weights is None:
            half_length = 0.5 * self.length
            lb = np.clip(x_center - half_length, 0.0, 1.0)
            ub = np.clip(x_center + half_length, 0.0, 1.0)
        else:
            if weights.ndim != 1 or weights.shape[0] != self.num_dim:
                raise ValueError(weights.shape)
            half_length = weights * self.length / 2.0
            lb = np.clip(x_center - half_length, 0.0, 1.0)
            ub = np.clip(x_center + half_length, 0.0, 1.0)
        return lb, ub

    def create_bounds(self, x_center) -> tuple:
        if (
            x_center.ndim != 2
            or x_center.shape[0] != 1
            or x_center.shape[1] != self.num_dim
        ):
            raise ValueError(x_center.shape)
        lb_1d, ub_1d = self._compute_bounds_1d(x_center[0])
        lb = lb_1d[None, :]
        ub = ub_1d[None, :]
        return lb, ub
